'**********************************************************************
'*  Name       : RX_buff.inc                                          *
'*  Author     : John Barrat with help from Darrel Taylor & Tim Box   *
'*  Date       : 20/04/05                                             *
'*  Version    : 1.0                                                  *
'*  Notes      : Requires Int_Ctrl-18F_P+.inc included in main prog.  *
'*             : 		  					  		   	       		  *
'**********************************************************************
' This program implements variable size circular transmit and receive buffers

' To Transmit:
' Set TB_Size to the desired size - must be a power of 2
' Variable TB_Len shows the number of bytes in the buffer
' HSerOutI [StrArray] to send a String Array to the buffer

' To Receive:
' Set RB_Size to the desired size - must be a power of 2 
' Variable RB_Len shows the number of bytes in the buffer
' Use RB_GetChar to get a byte from the buffer
'    The result will be in RXchar - Null if empty
' Set RB_MinMsgLen to minimum length before attempting to process the message
' 


Declare FSR_CONTEXT_SAVE = On  
'Device = 18F252

Symbol OERR	 = RCSTA.1			' RCSTA bit 1 OERR: Overrun Error bit
Symbol FERR	 = RCSTA.2			' RCSTA bit 2 FERR: Framing Error
Symbol CREN	 = RCSTA.4			' RCSTA bit 4 CREN: Continuous Receive Enable bit
Symbol RXEN	 = PIE1.5			' USART RX Interrupt Enable
Symbol RCIF	 = PIR1.5			' USART RX Interrupt Flag bit

Symbol TXIE	 = PIE1.4			' USART TX Interrupt Enable 
Symbol TXIF  = PIR1.4  			' USART TX Interrupt Flag bit
Symbol TXEN  = TXSTA.5    		' USART TX Enable (Usart 1)

Symbol RxTOut   5		   		' RX Timeout Value (2.5 Secs)
Symbol RB_Size	$20				' Size of Receive Buffer (must be a power of 2)
Symbol TB_Size  $20				' Size of Transmit Buffer (must be a power of 2)
Symbol TB_Mask = TB_Size - 1	'
Symbol RB_Mask = RB_Size - 1    '

Dim RXchar	As Byte             ' Receive Character
Dim RB_head	As Byte				' Pointer to start of receive buffer
Dim RB_Tail	As Byte				' Pointer to end of receive buffer
Dim RB_Len	As Byte             ' number of chars in receive buffer

Dim RB_OFL	As Bit				' 1 if Receive buffer overflows
Dim RB_FERR	As Bit				' 1 if USART reports Framing ERROR
Dim RB_CERR	As Bit				' 1 if Hex Conversion failed
Dim RB_ORun	As Bit				' USART Receive Overun

Dim TXChar  As Byte             ' Transmit Character
Dim TB_head	As Byte				' Pointer to start of transmit buffer
Dim TB_Tail	As Byte				' Pointer to end of transmit buffer
Dim TB_OFL  As Bit				' 1 if Transmit buffer overflows
Dim TB_Len  As Byte             ' number of chars in transmit buffer

Dim RBuff[RB_Size] As Byte		' The Receive buffer
Dim TBuff[TB_Size] As Byte		' The Transmit buffer

Dim TM_Byte	As Byte
Dim RXByte  As Word
Dim RXWord	As Word
Dim RxTOCtr	As Byte	   			' Timeout counter for Received Record
Dim AddressPTR As Word

GoTo Skip_Rbuff_subs

HISerOut MACRO  P1
#IF(PRM_COUNT != 1) 
   #ERROR "Incorrect number of parameters in HSerOutI macro call"
#ELSE
   #IF !(PRM_1 == 9)
       #ERROR "Param 1 should be a pointer to a string array"
   #ENDIF
; Set P1 address into FSR1
    Num_word P1, FSR1L
    GoSub HISerOut_Handler
#ENDIF
ENDM

' USART Initialisation
USARTBuffInit:
	Set CREN          		       	' Enable USART Continuous Receive
    Set TXEN		  		  	    ' Enable USART Transmit
	Clear RxTOCtr                   ' Reset Receive timeout
	GoSub RB_ClearBuff              ' Clear the receive buffer
#IFDEF TX_HANDLER
' Clears the TX buffer pointers
TB_ClearBuff:                       
	TB_Len = 0: TB_Tail = 0: TB_head = 0
	Clear TXIE	                    ' Disable TX interrupt
TB_ClearErr:                        ' Clear TX Error values
	Clear TB_OFL
#ENDIF
Return

' Clears the Receive buffer pointers 
RB_ClearBuff:
	RB_Len = 0: RB_Tail = 0: RB_head = 0
RB_ClearErr:              	 		' Clear RX Error values
	Clear RB_FERR
	Clear RB_OFL
	Clear RB_ORun
Return

'USART Transmit interrupt handler **************************************************************
INT#STARTH:
#IFDEF TX_HANDLER
TransmitUSART:
	TXREG = TBuff[TB_head]          ' Put Char at Head of the FIFO into the USART TX register
	Inc TB_head: Dec TB_Len         ' Reposition Head of FIFO and Dec FIFO Length
	TB_head = TB_head & TB_Mask 	' roll over to 0 if end reached
	If TB_Len = 0 Then     		   	' If Empty
		Clear TXIE  	  	  		' switch off interrupt
#IFDEF PUT_TX_RECORD
	    ADDImmd Put_Record          ' See if there is another record to get
#ENDIF
    EndIf
GoTo INT_Code_Cont					' Return to Interrupt router
#ENDIF

'USART Receive Interrupt handler ***************************************************************
ReceiveUSART:
	If OERR = 1 Then			    ' RCSTA bit 1 OERR: Overrun Error bit 
		Clear CREN: Set CREN 		' Clear OERR bit by turning CREN off then back on again.
		Set RB_ORun                 ' Set the Over-run flag
	EndIf
	RB_FERR = FERR	   	 		    ' Save Framing Error status
	If RB_Len >= RB_Size Then RBuffErr	' If RX buffer not Full
    RBuff[RB_Tail] = RCREG		    ' Add a char to the Receive Buffer
    Inc RB_Tail: Inc RB_Len		    ' Increment end pointer and length
    RB_Tail = RB_Tail & RB_Mask     ' Roll back to start of buffer if reached end
    GoSub StartRXTmr
#IFDEF GET_RX_RECORD
    If RB_Len > RB_MinMsgLen Then ADDImmd Get_Record' Time to look for a complete record
#ENDIF
GoTo INT_Code_Cont                	' return from the interrupt router

RBuffErr:
    Set RB_OFL						' Receive Buffer overflow
GoTo INT_Code_Cont                	' return from the interrupt router
INT#ENDH: 

'END - USART Receive Interrupt handler *********************************************************
SMALL_MICRO_MODEL = On
' RX Timer Subroutines *************************************************************************
RXTmr:                                         ' RX Timout
    If RxTOCtr = 0 Then REMTask RXTmr,Rate2    ' If set to 0 then remove timeout timer
    Dec RxTOCtr								   ' Decrement the Timout counter
    If RxTOCtr = 0 Then						   ' Must have lost contact
       GoSub RB_ClearBuff					   ' Reset the RX Buffer pointers
	   GoTo SendTOErr:						   ' and Return through send TimeOut error
    EndIf 
Return

StopRXTmr: 		  						       ' Stops the RX Timer prematurely
   RxTOCtr = 0
Return

StartRXTmr:
   If RxTOCtr = 0 Then ADDTask RXTmr, Rate2
   RxTOCtr = RxTOut
Return

' End of RX Timer routines *********************************************************************

' Macro Handler - Sends a string of characters passed in array to TX USART buffer
HISerOut_Handler:
    TXChar = POSTINC1                   	 ' Get first character
	While TXChar > Null                   	 ' Loop until we detect a Null
	    While TB_Len >= TB_Size: Wend 	     ' Hang around until we can load character
		TBuff[TB_Tail] = TXChar		   	  	 ' Add a char to the Transmit Buffer
		Clear TXIE                           ' Disable TX interrupt while adjusting pointers
		Inc TB_Tail: Inc TB_Len		   	  	 ' Increment end pointer and length	
		TB_Tail = TB_Tail & TB_Mask          ' Roll over to start if end of FIFO
		Set TXIE                             ' enable interrupt now we've re adjusted pointers
	    TXChar = POSTINC1                    ' Get next character
	Wend                                     ' Loop
Return

RB_GetChar:									' Get a char from the Receive Buffer
	RXchar = Null                           ' Set Null return char in case Len = 0
	If RB_Len > 0 Then                      ' If buffer contains characters
		RXchar = RBuff[RB_head]             ' Load the character at the top of the buffer
        Clear RXEN						  	' Mask Receive interrupt while changing buffer
		Inc RB_head: Dec RB_Len             ' Reposition the pointers
		RB_head = RB_head & RB_Mask         ' and roll over if we have reached the end
	    Set RXEN							' Re-enable Receive interrupt
	EndIf
Return

' Returns first char in buffer without modifying buffer pointers
RB_GetFirstChar:
	RXchar = 0
	If RB_Len > 0 Then RXchar = RBuff[RB_head]
Return

' Returns the last received byte without modifying buffer pointers
RB_GetLastChar:
    RXchar = 0
    If RB_Len > 0 Then RXchar = RBuff[RB_Tail - 1]
Return

' Gets a 4 character Hex word from the buffer - Sets RB_CERR if Character is not a Hex number
RB_GetHex4:
    GoSub RB_GetHex2
    If RB_CERR = 1 Then Return
    RXWord.HighByte = RXByte
    GoSub RB_GetHex2
    RXWord.LowByte  = RXByte
Return

' Gets a 2 character Hex Byte from the buffer - Sets RB_CERR if Character is not a Hex number
RB_GetHex2:
	RXByte = 0
	GoSub RB_HexConv
	If RB_CERR = 1 Then Return
	GoSub RB_HexConv
Return

' Converts a single Character into a Hex number if character represents a Hex value
RB_HexConv:	
    RB_CERR = 0: TM_Byte = 99: GoSub RB_GetChar
	TM_Byte = LookDown RXchar,["0123456789ABCDEF"]
	If TM_Byte = 99 Then RB_CERR = 1
    RXByte = RXByte << 4
	RXByte = RXByte | TM_Byte
Return
SMALL_MICRO_MODEL = OFF 

Skip_Rbuff_subs:
