'****************************************************************
'*  Name    : TASKMAN_INC.pbp                                   *
'*  Author  : John Barrat                                       *
'*  Notice  : Copyright (c) 2004 John Barrat                    *
'*          : All Rights Reserved                               *
'*  Date    : 26/10/2004                                        *
'*  Version : 1.0                                               *
'*  Notes   : Requires the following includes:                  *
'*          :    CON_INC.Bas                                    *
'****************************************************************

' Add the following to the user.mcd file

'KEY ("ADDTask")
'   DeviceCore = [dc16]
'   Highlighter = Keyword 
'   Hint = "<b>ADDTask</b> <i>[<b>Task</b> As Label, <b>Rate</b> As Const], <b>Result</b> As Byte</i>"
'ENDKEY

'KEY ("ADDImmd")
'   DeviceCore = [dc16]
'   Highlighter = Keyword 
'   Hint = "<b>ADDImmd</b> <i>[<b>Task</b> As Label</i>"
'ENDKEY

'KEY ("REMTask")
'   DeviceCore = [dc16]
'   Highlighter = Keyword
'   Hint = "<b>REMTask</b> <i>[<b>Task,</b> As Label, <b>Rate</b> As Const]</i>, <b>Result</b> As Byte</i>"
'ENDKEY

'KEY ("POSTask")
'   DeviceCore = [dc16]
'   Highlighter = Keyword
'   Hint = "<b>POSTask</b> <i>[<b>FirstTask</b> As Label, <b>SecondTask</b> As Label, <b>Rate</b> As Const]</i>, <b>Result</b> As Byte</i>"
'ENDKEY

'KEY ("REPTask")
'   DeviceCore = [dc16]
'   Highlighter = Keyword
'   Hint = "<b>REPTask</b> <i>[<b>CurrentTask</b> As Label, <b>NewTask</b> As Label, <b>Rate</b> As Const]</i>, <b>Result</b> As Byte</i>"
'ENDKEY


'Device = 18F252
Symbol MaxImmdTasks  = ImmdTasks                      ' Maximum number of Interrupt tasks
Symbol ImmdMask		 = ImmdTasks - 1                  ' Set the FIFO rollover mask
Symbol MaxRate1Tasks = Rate1Tasks + MaxImmdTasks      ' Maximum number of tasks every 500mSec
Symbol MaxRate2Tasks = Rate2Tasks + MaxRate1Tasks     ' Maximum number of tasts every Second                
Symbol MaxRate3Tasks = Rate3tasks + MaxRate2Tasks     ' Maximum number of tasks every Minute               
Symbol MaxRate4Tasks = Rate4tasks + MaxRate3Tasks     ' Maximum number of tasks every Hour				
Symbol MaxCoDTasks   = CoDTasks   + MaxRate4Tasks     ' Maximum number of Tasks at Change of Day

' Internal Variables
Dim MaxTasks     As Byte
Dim Tasks[MaxCoDTasks] As Word	  ' Task List
Dim TaskPTR      As Byte	  	  ' Pointer to Current Task
Dim TM_Int       As Byte          ' GP Int Var
Dim TM_Int2      As Byte          ' GP Int Var
Dim Address      As Word          ' Address of Task
Dim Address1     As Word
Dim TM_Len       As Byte   		  ' Length of the Immediate Task List
Dim TM_Head      As Byte   		  ' Pointer to the Start of the Immediate Task List
Dim TM_Tail      As Byte   		  ' Pointer to the End of the Immediate Task list

' External Control Variables
Dim ListFull     As Bit
Dim AlreadyThere As Bit
Dim RptTsks	     As Bit	  		  ' Switch on/off Task list reporting.
Dim RptActTsks	 As Bit			  ' Switch on/off Active Task reporting

' External Constants:
Dim Rate1        As 1
Dim Rate2        As 2
Dim Rate3        As 3
Dim Rate4        As 4
Dim CoD          As 5
Dim RateAll      As 0

GoTo Skip_TM

'ADD a Task to the list-------------------------------------------------------------------------
' This Macro adds a task to the Task list
' Syntax ADDTask Label, Rate
' Where - Label is the label of the Task to added
'       - Rate is the repetition rate at which the task is to be run 
' It will set ListFull and AlreadyThere booleans before returning.

ADDTask MACRO P1, P2
#IF(PRM_COUNT != 2) 
   #ERROR "ADDTask - Incorrect number of parameters"
#ELSE
   #IF !(PRM_1 == LABEL)
       #ERROR "ADDTask Param 1 must be a label"
   #ENDIF
   #IF ! (PRM_2 == NUM8)
       #ERROR "ADDTask Param 2 must be an 8 bit number"
   #ENDIF
   #IF (PRM_1 == LABEL) && (PRM_2 == NUM8)
        NUM_BYTE P2, TM_Int
        NUM_WORD P1, Address
        GoSub ADDTask_Handler
   #ENDIF
#ENDIF
ENDM


#IFDEF INT_ENABLED
'ADD an Interrupt Task to the list--------------------------------------------------------------
' This Macro adds a high Priority task to the Task list
' The task will execute immediately the current task has completed
' and will be executed once only
' Syntax ADDImmediate Label
' Where - Label is the label of the Task to added
' It will set ListFull boolean before returning.
ADDImmd MACRO P1
#IF (PRM_COUNT !=1)
    #ERROR "ADDImmd - Incorrect number of parameters"
#ELSE
    #IF !(PRM_1 == LABEL)
        #ERROR "ADDImmd Param must be a label"
    #ELSE
        NUM_WORD P1, Address
        GoSub ADDImmed_Handler
    #ENDIF
#ENDIF
ENDM
#ENDIF

'REMove a Task from the list--------------------------------------------------------------------
' This Macro removes a task from the Task list
' Syntax REMTask Label, Rate
' Where - Label is the label of the Task to removerd
'       - Rate is the repetition rate at which the task was being run 
' It will set ListFull and AlreadyThere booleans before returning.
REMTask MACRO P1 ,P2
#IF(PRM_COUNT != 2) 
   #ERROR "REMTask - Incorrect number of parameters"
#ELSE
   #IF !(PRM_1 == LABEL)
       #ERROR "REMTask Param 1 should be a label"
   #ENDIF
   #IF ! (PRM_2 == NUM8)
       #ERROR "REMTask Param 2 should be an 8 bit number"
   #ENDIF
   #IF (PRM_1 == LABEL) && (PRM_2 == NUM8)
        NUM_BYTE P2, TM_Int
        NUM_WORD P1, Address
        GoSub REMTask_Handler
   #ENDIF
#ENDIF
ENDM

' Replaces an existing task with another task
' Syntax REPTask Label1, Label2, Rate
' Where - Label 1 is the label of the existing task, 
' 		  Label 2 is the replacement task
'		  and Rate is the repetition rate the task existing task was being run

REPTask MACRO P1, P2, P3
#IF(PRM_COUNT != 3) 
   #ERROR "Incorrect number of parameters in REPTask macro call"
#ELSE
   #IF !(PRM_1 == LABEL)
       #ERROR "Param 1 should be a label"
   #ENDIF
   #IF !(PRM_2 == LABEL)
       #ERROR "Param 2 should be a label"
   #ENDIF
   #IF !(PRM_3 == NUM8)
       #ERROR "Param 2 should be an 8 bit number"
   #ENDIF
   #IF (PRM_1 == LABEL) && (PRM_2 == LABEL) && (PRM_3 == NUM8)
       NUM_BYTE P3, TM_Int
       NUM_WORD P1, Address
       NUM_WORD P2, Address1
       GoSub REPTask_Handler
   #ENDIF
#ENDIF
ENDM

'POSitions a Task relative to another task in the list------------------------------------------
' This macro repositions two tasks relative to each other.
' Syntax POSTask FirstLabel, SecondLabel, Rate
'  Tasks must be present in the list
' will return with AlreadyThere false if either task is not found.  
' If the order is incorrect will swap the positions of the two tasks.
' When called it should pass the following parameters:
'   FirstAddress        The address of the first Task Label
'   SecondAddress       The address of the second Task Label
'   Rate                The rate the tasks were activated

POSTask MACRO P1 ,P2, P3
#IF(PRM_COUNT != 3) 
   #ERROR "Incorrect number of parameters in REMTask macro call"
#ELSE
   #IF !(PRM_1 == LABEL)
       #ERROR "Param 1 should be a label"
   #ENDIF
   #IF !(PRM_2 == LABEL)
       #ERROR "Param 2 should be a label"
   #ENDIF
   #IF ! (PRM_3 == NUM8)
       #ERROR "Param 2 should be an 8 bit number"
   #ENDIF
   #IF (PRM_1 == LABEL) && (PRM_2 == LABEL) && (PRM_3 == NUM8)
       NUM_BYTE P3, TM_Int
       NUM_WORD P1, Address
       NUM_WORD P2, Address1
       GoSub POSTask_Handler
   #ENDIF
#ENDIF
ENDM

' Clears down the Tasks Array
SMALL_MICRO_MODEL = On
TaskMan_Init:
    TaskPTR = 0
    MaxTasks = MaxCoDTasks
' Clears all the tasks in the Task list indicated by Set Rate
ClrAllTasks:
 	Clear Tasks
    TM_Head = 0: TM_Tail = 0: TM_Len = 0   	   ' Reset Immediate FIFO pointers and Length to 0
    TM_Int = "C"                           	   ' Set result
#IFDEF TASK_DEBUG
	RptTsks = False 			  		   	   ' Set Task Reporting to Off
	RptActTsks = False            		   	   ' Set Active Task Reporting to Off
#ENDIF
    GoTo CommonReturn

' This routine points the Task Manager to the set of task rates
' passed with the call  -  Example GoSub Rate [Rate1]                                                                           
Rate:                                    		' External Rate command
   Pop TM_Int                              		' retreive the Rate into TM_Int
RateInt:                                   		' Internal Rate command - TM_Int already set
   TaskPTR  = LookUp TM_Int, [0, MaxImmdTasks,  MaxRate1Tasks, _            
 	                          MaxRate2Tasks, MaxRate3Tasks, MaxRate4Tasks]' Set up the TaskPtr
   MaxTasks = LookUp TM_Int, [MaxCoDTasks,   MaxRate1Tasks, MaxRate2Tasks, _  
                              MaxRate3Tasks, MaxRate4Tasks, MaxCoDTasks]  ' Set up the MaxTask
Return

TaskHandler:
' Task handler will run through all the tasks in the task list for the current Task Rate
' it will return on the first task with an address of 0 or if the task count has reached 
' the max for the current task rate.
GoSub Rate                                  	' Get ptr to start of task list for current rate
Repeat
    Address = Tasks[TaskPTR]                    ' Get the Address of the next task
    If Address = 0 Then Break         			' a 0 address means no further tasks
    Push TaskPTR                                ' save TaskPTR in case current code alters it
	GoSub Execute                               ' Start the task at Address
    Pop TaskPTR                                 ' restore TaskPTR
	Inc TaskPTR                       			' move on down the task list
Until TaskPTR = MaxTasks                    	' and try again
Return

#IFDEF INT_ENABLED
Immediate_Tasks:									
' Immediate_Tasks runs through all Immediate Tasks in task list. Each will only execute once.
' Addresses of Immediate Tasks held in a circular buffer - once executed they must be re-added 
' to the buffer in order to execute again. 
While TM_Len > 0 
	Address = Tasks[TM_Head]                    ' Get Address from the Buffer
#IFDEF TASK_DEBUG
    TaskPTR = 0									' TaskPTR set to indicate Immediate task run
#ENDIF
	GoSub Execute                               ' Run task at that address
	Clear Tasks[TM_Head]                        ' Clear head to indicate it has been executed
	Inc TM_Head: Dec TM_Len                     ' Incr pointer to the start of buffer dec Length
	TM_Head = TM_Head & ImmdMask                ' Roll over to 0 if end of buffer
Wend											' Loop until all entries done
#ENDIF
Return                                      	' return to scheduler

Execute:
' GoTo the address of a Task from a word variable and Return via Task to Handler
#IFDEF TASK_DEBUG
	GoSub SendActTask							' Report the task we are about to run
#ENDIF   
ASM
    movff Address + 1, PCLATH                   ; Set PCH via PCLATH
	movff Address, WREG                         ; Can't write direct to PCL with movff
    movwf PCL                                   ; Writing to PCL will also load PCLath
ENDASM

#IFDEF INT_ENABLED
'ADDImmediate ----------------------------------------------------------------------------------
' Add Immediate Task to Task List.  Sets ListFull if there is no room in buffer
' The Immediate Task List is organised as a FIFO. Tasks will be removed from list on execution
ADDImmed_Handler:
	Clear AlreadyThere                       	' OK so reset flag
    If Address = Tasks[TM_Head] Then 
       Set AlreadyThere                      	' Previous entry not processed set AlreadyThere 
	   Return                                   ' and return
    EndIf	   
	If TM_Len < MaxImmdTasks Then			    ' If Immediate buffer not Full
	   Tasks[TM_Tail] = Address				    ' Add Immediate task Address to the Task list
	   Inc TM_Tail: Inc TM_Len	                ' point to new end of buffer & adjust the length
	   TM_Tail = TM_Tail & ImmdMask             ' Roll over to 0 at end of buffer
       Clear ListFull: TM_Int = "A"           	' Set the status flags	   
	Else
	   Set ListFull								' Buffer overflow
	   GoTo ActFail                             ' Failed to Add Immediate Routine
	EndIf
	#IFDEF TASK_DEBUG
	TaskPTR = 0	  								' TaskPTR set to Immediate Added
	#ENDIF
GoTo CommonReturn
#ENDIF

' Restores Task Pointer ------------------------------------------------------------------------

'FIND a task in the list -----------------------------------------------------------------------
FINDTask:
   For TM_Int = TaskPTR To MaxTasks - 1         ' Scan through the List
      If Tasks[TM_Int] = Address Then           ' if Task is in the list?
         Set AlreadyThere                    	' indicates that the Task was found
         Return                      		    ' returns with flag set and position in TM_Int
      EndIf
   Next TM_Int
   Clear AlreadyThere
Return

' ADDTask --------------------------------------------------------------------------------------
' This routine adds a task to the Task list 
' It will set ListFull and AlreadyThere booleans before returning.
' It will return the result of the Addition.
ADDTask_Handler:
'HSerOut ["TM_Int ", Dec TM_Int, CR, LF]
    GoSub RateInt                               ' Set up the Tasklist Pointers
    GoSub FINDTask                              ' Look for the task in the list
    If AlreadyThere = False Then                ' Not there so 
		For TM_Int2 = TaskPTR To MaxTasks - 1   ' Look for the first empty slot
            If Tasks[TM_Int2] = 0 Then          ' Found first empty space
                Tasks[TM_Int2] = Address        ' add in the new address and
                Clear ListFull                	' and set success boolean
                GoTo ADDTaskDone                ' exit outer loop
            EndIf
        Next TM_Int2                            ' keep looking for empty slot
        Set ListFull                         	' No room for the Task
    EndIf
    GoTo ActFail                                ' So return through Error
ADDTaskDone:
    TM_Int = "A"
    GoTo CommonReturn

'POSitions a Task relative to another task in the list------------------------------------------
' Positions the tasks in the order requested.  Tasks must already be present in the list
POSTask_Handler:
'    Pop Address, Address1                       ' Retrieve both Addresses
    GoSub RateInt                               ' Set up the Tasklist Pointers
    GoSub FINDTask                              ' Find Address of first task
    If AlreadyThere = False Then ActFail   		' Can't find first address so exit
    TM_Int2 = TM_Int                            ' Save position in TM_Int2
    Swap Address, Address1: GoSub FINDTask      ' Find the Address of second task
    If AlreadyThere = False Then  ActFail   	' Can't find it so exit
    ' TM_Int2 holds position of Address and TM_Int holds position of Address1
    If TM_Int < TM_Int2 Then  					' Test for correct order
        Tasks[TM_Int]  = Address1               ' Order incorrect so swap the order
        Tasks[TM_Int2] = Address                ' of the addresses
    EndIf
    TM_Int = "P"                                ' Set return to "P" to confirm re-order
    GoTo CommonReturn

'REPlace an existing task with a different task-------------------------------------------------
' This routine locates the task defined in Address and replaces it with the task in Address1
REPTask_Handler:
'    Pop Address, Address1                       ' Retrieve both Addresses
    GoSub RateInt                               ' Set up the Tasklist Pointers
    GoSub FINDTask                              ' Find Address of first task
    If AlreadyThere = False Then ActFail   		' Can't find first address so exit
    Tasks[TM_Int] = Address1                    ' Replace the task
    TM_Int = "S"                                ' Set return to "S" to confirm replacement
	GoTo CommonReturn
    
'REMove a Task from the list--------------------------------------------------------------------
' This routine removes a task from the current set Rate Task
REMTask_Handler:
    GoSub RateInt                               ' Set up the Tasklist Pointers
    GoSub FINDTask                              ' Find Address of task
    If AlreadyThere = False Then ActFail        ' Task not found in the List
    Tasks[TM_Int] = 0                           ' Found it so remove task
    TM_Int2 = TM_Int                         
    For TM_Int = TM_Int2 To MaxTasks - 1        ' Shift remaining tasks UP 1 slot
        Tasks[TM_Int] = Tasks[TM_Int + 1]       ' copy next Slot to current Slot
        If Tasks[TM_Int + 1] = 0 Then           ' if next Slot is empty?
            GoTo ShiftDone                      ' shift is finished
        EndIf
    Next TM_Int
ShiftDone:
   Tasks[MaxTasks - 1] = 0                      ' Clear the last slot in the List
   Clear ListFull                            	' show that there's room in the List
    TM_Int = "R"
GoTo CommonReturn
SMALL_MICRO_MODEL = OFF

'Common exit routine for ADD, REM, POS, REP and Clear-------------------------------------------
ActFail:
    TM_Int = "E"
CommonReturn:
#IFNDEF TASK_DEBUG
    Return
#ELSE

' Debug code - prints out the addresses held in the task list.**********************************
SMALL_MICRO_MODEL = On
If RptTsks = False Then Return 
If TM_Int = "E" Then Return
ReportTasks:
' Outputs 'Active Tasks Addresses' on the Serial port for most recently changed Rate in the form
'   TaskPTR Value, AlreadyThere, ListFull, Action (A=Add, R=remove, P=reposition, S=Replace)
'   Dec Address, Dec Address ....... CR LF'
   MsgArry = "^RTK," + Str$(DEC2 TaskPTR) + "," + Str$(BIN1 AlreadyThere) + "," + Str$(BIN1 ListFull) + "," + TM_Int + "," + Str$(HEX4 Address)
   If TM_Int = "S" Or TM_Int = "P" Then MsgArry = MsgArry  + "," + Str$(HEX4 Address1)
   GoTo SendDebugMsg

SendActTask:
' Outputs the Current Rate in task list and address of the task about to be executed
   If RptActTsks = False Then Return
   MsgArry = "^RAT," + Str$(DEC2 TaskPTR) + "," + Str$(HEX4 Address)
   GoTo SendDebugMsg

' This routine returns a list of task pointer start addresses for each task rate
' It assumes that task rate 1 starts at 0.  Use this to determine the start locations in the
' list for each Task Rate and the max number of tasks for each rate.
SendRateLocs:
   MsgArry = "^RTR," + Str$(DEC2 MaxImmdTasks) + "," + Str$(Dec MaxRate1Tasks) + "," + Str$(Dec MaxRate2Tasks) + "," + Str$(Dec MaxRate3Tasks) + "," + Str$(Dec MaxRate4Tasks) + "," + Str$(Dec MaxCoDTasks)	     
SendDebugMsg:
SMALL_MICRO_MODEL = Off
   GoTo SendMessage

'This routine sends out a snapshot of the whole task table. Use this to prime the initial state
SendAllTasks:
SMALL_MICRO_MODEL = On
TM_Int = "I": AlreadyThere = 0: ListFull = 0
TaskPTR = ImmdTasks
Repeat
   Address = Tasks[TaskPTR]
   If Address > 0 Then GoSub ReportTasks
   Inc TaskPTR
Until TaskPTR = MaxCoDTasks -1 
Return
SMALL_MICRO_MODEL = Off

#ENDIF  
Skip_TM:
